#include "progresswindow.h"
#include "settings.h"
#include "messages.h"

#include <cstdlib>

#define DEBUG( arg... ) do { if( m_pcSettings->GetDebugMode() ) printf( arg ); } while(0)

/** \brief Get Resource from String.
 * This wrapper function is used to help get the resources for the images
 * within the file.
 *
 * \param fileName The file that we're getting the resource image of.
 *
 * \return A pointer to the BitmapImage corresponding to the resource image.
 */
os::BitmapImage *getResource(const os::String& fileName)
{
	os::Resources cCol( get_image_id() ); // For getting image streams
	os::ResStream *pcStream = cCol.GetResourceStream( fileName );
	os::BitmapImage *pcIcon = new os::BitmapImage( pcStream );
	return pcIcon;
}

/** \brief Converts a number into a human-readable size (eg 2.4MB)
 */
String HumanReadableSize( ssize_t nSize )
{
	String cTmp;

	if( nSize < (2<<10) ) { /* 0B - 1KB */
		cTmp = cTmp.Format( "%liB", nSize );
	} else if( nSize < ((uint64)1<<20) ) { /* 1KB - 1MB */
		cTmp = cTmp.Format( "%.1fKB", (float)nSize / ((uint64)1<<10) );
	} else if( nSize < ((uint64)1<<30) ) { /* 1MB - 1GB */
		cTmp = cTmp.Format( "%.2fMB", (float)nSize / ((uint64)1<<20) );
	} else if( nSize < ((uint64)1<<40) ) { /* 1GB - 1TB */
		cTmp = cTmp.Format( "%.2fGB", (float)nSize / ((uint64)1<<30) );
	} else { /* > 1TB */
		cTmp = cTmp.Format( "%.2fTB", (float)nSize / ((uint64)1<<40) );		
	}
	return( cTmp );
}


/** \brief ProgressListView class
    We only need to subclass this to catch right-clicks for the context menu.
*/

class ProgressListView : public os::ListView
{
public:
	ProgressListView( const os::Rect& cFrame, const os::String& cTitle, uint32 nModeFlags, uint32 nResizeMask )
		: os::ListView( cFrame, cTitle, nModeFlags, nResizeMask )
	{ }
	
	~ProgressListView()
	{ }
	
	void MouseDown( const os::Point& cPosition, uint32 nButton );
};

/** \brief Catch right-clicks and display a context menu.
 */
void ProgressListView::MouseDown( const os::Point& cPosition, uint32 nButton )
{
	if( nButton == 2 ) {
		os::ListViewStringRow* pcRow = (os::ListViewStringRow*)GetRow( cPosition );
		if( pcRow != NULL ) {
			os::Point cTmp = cPosition;
			ConvertToScreen( &cTmp );
			int nID = atoi( pcRow->GetString( ID_INDEX ).c_str() );
			((ProgressWindow*)GetWindow())->OpenContextMenu( cTmp, nID );
			return;
		}
	}
	ListView::MouseDown( cPosition, nButton );
}



/** \brief Progress Window.
 *
 * \todo Must change the panels so that when the window is resized,
 * the ListView is resized as well.
 *
 * \todo Need to set up the ListView so that it's scroll bars will
 * become visible when necessary.
 */
ProgressWindow::ProgressWindow(MainWindow *mainWindow, AppSettings* pcSettings)
	: os::Window( os::Rect( 0, 0, 500, 200 ), "progress_wnd", "Progress Window" )
{
	m_pcMainWindow = mainWindow; // Keep track of the main window so we can send messages to it.
	m_pcSettings = pcSettings;
	
	os::LayoutView* pcView = new os::LayoutView( GetBounds(), "layout_view" );
	AddChild( pcView );
	
	/* Tool Bar */
	os::ToolBar *myToolBar = new os::ToolBar(os::Rect(0, 0, GetBounds().Width(), 40), "toolbar");
	myToolBar->AddButton("play_button", "", getResource("play_button.png"), new os::Message(M_RESUME));
	myToolBar->AddButton("pause_button", "", getResource("pause_button.png"), new os::Message(M_PAUSE));
	myToolBar->AddButton("stop_button", "", getResource("stop_button.png"), new os::Message(M_CANCEL));
	pcView->AddChild(myToolBar);
	
    /* List View */
	m_pcListView = new ProgressListView( os::Rect(0, 40, GetBounds().Width(), GetBounds().Height() - 20), "ListView", os::ListView::F_MULTI_SELECT | os::ListView::F_RENDER_BORDER, os::CF_FOLLOW_ALL );

	// First Create the Columns
	int nUnitWidth = (int)GetBounds().Width();
	nUnitWidth /= 11;
	m_pcListView->InsertColumn("Status", nUnitWidth*2, STATUS_INDEX);
	m_pcListView->InsertColumn("Source", nUnitWidth*3, SERVER_INDEX);
	m_pcListView->InsertColumn("Destination", nUnitWidth*3, FILE_INDEX);	
	m_pcListView->InsertColumn("Total Transferred", nUnitWidth, TRANSFERRED_INDEX);
	m_pcListView->InsertColumn("File Size", nUnitWidth, FILE_SIZE_INDEX);
	m_pcListView->InsertColumn("ID", nUnitWidth, ID_INDEX);

	pcView->AddChild(m_pcListView);

		
	/* Status Bar */
    m_pcStatusBar = new os::StatusBar(os::Rect(0,GetBounds().Height()-20,GetBounds().Width(),GetBounds().Height()),"",os::CF_FOLLOW_LEFT | os::CF_FOLLOW_RIGHT | os::CF_FOLLOW_BOTTOM);
    m_pcStatusBar->AddPanel("Status Panel", "2 Downloads / 0 Uploads / 0 in Queue");
    m_pcStatusBar->AddPanel("Percentage Panel", "Percent Complete: 5.86%");
	pcView->AddChild(m_pcStatusBar);

	/* Set up the initial status. */
	UpdateTransfers(NULL);	//&testMessage);

	/* Set Icon */
	os::Resources cCol( get_image_id() );
	os::ResStream *pcStream = cCol.GetResourceStream( "icon48x48.png" );
	os::BitmapImage *pcIcon = new os::BitmapImage( pcStream );
	SetIcon( pcIcon->LockBitmap() );
	delete( pcIcon );
}

ProgressWindow::~ProgressWindow()
{

	/*
	delete(m_pcRoot);
	delete(m_pcStatusBar);
	delete(m_pcListView);
	*/
}


/** \brief Update Transfers.
 * This function takes in the updated list of files being transfered and clears the list view. It then
 * updates it with the new list of transfer files.
 *
 * \param psUpdate The pointer to the update structure for the new list of transfers.
 *
 */
void ProgressWindow::UpdateTransfers(Message *pcMessage)
{
	int totalDownloads = 0;
	int totalUploads = 0;
	int totalQueued = 0;
	double percentComplete = 0.0;
	os::String result; // Used to store the output string
	
	// First clear the old transfers.
	m_pcListView->Clear();
	
	// Update the list of items.
	if ( pcMessage != NULL )
	{
		std::vector< QueueItem >* pacNodes;
		if( pcMessage->FindPointer( "data", (void**)&pacNodes ) != 0 ) {
			printf( "ProgressWindow::UpdateTransfers: Got message without data pointer!\n" );
			return;
		}
		
		if( pacNodes != NULL )
		{
			std::vector< QueueItem >::iterator i = pacNodes->begin();
			while( i != pacNodes->end() )
			{
				/* Create a list row and populate it with data */
				os::ListViewStringRow* pcRow = new os::ListViewStringRow();
				
				if( i->eStatus == NODE_QUEUED ) {
					pcRow->AppendString( "Queued" );
				} else if( i->eStatus == NODE_PAUSED ) {
					pcRow->AppendString( "Paused" );
				} else if( i->eStatus == NODE_RUNNING ) {
					if( i->eType == NODE_DOWNLOAD ) {
						pcRow->AppendString( "Downloading" );
					} else if ( i->eType == NODE_UPLOAD ) {
						pcRow->AppendString( "Uploading" );
					} else {
						pcRow->AppendString( "Unknown" );
						printf( "ProgressWindow::UpdateTransfers: Got unknown transfer type!\n" );
					}
				} else {
					pcRow->AppendString( "Unknown" );
					printf( "ProgressWindow::UpdateTransfers: Got unknown transfer status!\n" );
				}
				
				if( i->eType == NODE_DOWNLOAD ) {
					pcRow->AppendString( i->zRemotePath );
					pcRow->AppendString( i->zLocalPath );
					if( i->nTotalDownload > 0 ) {
						pcRow->AppendString( HumanReadableSize( i->nTotalDownload ) );
					} else {
						pcRow->AppendString( "" );
					}
				} else {
					pcRow->AppendString( i->zLocalPath );
					pcRow->AppendString( i->zRemotePath );
					if( i->nTotalUpload > 0 ) {
						pcRow->AppendString( HumanReadableSize( i->nTotalUpload ) );
					} else {
						pcRow->AppendString( "" );
					}
				}
				if( i->nTotalSize > 0 ) {
					pcRow->AppendString( HumanReadableSize( i->nTotalSize ) );
				} else {
					pcRow->AppendString( "" );
				}

				result = result.Format( "%d", i->nID );
				pcRow->AppendString( result );
			
				m_pcListView->InsertRow(pcRow);
				i++;
			}
			delete( pacNodes );
		}
		
		pcMessage->FindInt32( "total_downloads", &totalDownloads );
		pcMessage->FindInt32( "total_uploads", &totalUploads );
		pcMessage->FindInt32( "total_queued", &totalQueued );
	}
		
	// Once done, update the status bar with the correct transfer information.
    m_pcStatusBar->SetText("Status Panel", result.Format("%d Downloads / %d Uploads / %d in Queue", totalDownloads,
    	totalUploads, totalQueued));
    m_pcStatusBar->SetText("Percentage Panel", result.Format("Percent Complete: %.2g", percentComplete));
    /* TODO: total percent complete, individual percent complete */

#ifdef DEBUG_PW
	DEBUG("Updated the list\n");
#endif /* DEBUG_PW */
}


/** \brief Window Message Handler.
 * Handles messages sent to the window.
 *
 * \param pcMessage The messages being sent.
 */
void ProgressWindow::HandleMessage( os::Message* pcMessage )
{
	switch( pcMessage->GetCode() )
	{
		/* Handle Messages for Updating the . */
		case M_PW_UPDATE:
		{
			UpdateTransfers( pcMessage );
		} break;
		
		/* Handle messages from the buttons. */
		case M_PAUSE:
		{
			SendPauseMessage( pcMessage );
			break;
		}

		case M_RESUME:
		{
			SendResumeMessage( pcMessage );
			break;
		}

		case M_CANCEL:
		{
			SendCancelMessage( pcMessage );
			break;
		}

		/* Close the window. */
		case M_APP_QUIT:
		{
			PostMessage( os::M_QUIT );
			break;
		}
	}
}


/** Send a pause message for the current download to the main window.
 * \param pcMsg: The message from the GUI.
 *   If this is from the context menu, it contains a transfer id as the parameter 'id'.
 */
void ProgressWindow::SendPauseMessage( os::Message* pcMsg )
{
	os::Message cMessage( M_PW_PAUSE );
	int nID;
	if( pcMsg->FindInt32( "id", &nID ) == 0 ) {
		cMessage.AddInt32( "nodeID", nID );
	}
	// Go through each selected item (i is < 0 if nothing selected).
	for (int i = m_pcListView->GetFirstSelected(); i >= 0 && i <= m_pcListView->GetLastSelected(); i++)
	{
		if( m_pcListView->IsSelected( i ) ) {
			os::ListViewStringRow *row = (os::ListViewStringRow *) m_pcListView->GetRow(i);
			os::String cStr = row->GetString(ID_INDEX);
			int nodeID = atoi(cStr.c_str());

			cMessage.AddInt32("nodeID", nodeID);

			DEBUG("ProgressWindow: sending Pause Message for node %d\n", nodeID);
		}
	}
	m_pcMainWindow->PostMessage(&cMessage, m_pcMainWindow); // Send the message to the main window
	return;
}


/** Send a resume message for the current download to the main window.
 */
void ProgressWindow::SendResumeMessage( os::Message* pcMsg )
{
	os::Message cMessage( M_PW_RESUME );
	int nID;
	if( pcMsg->FindInt32( "id", &nID ) == 0 ) {
		cMessage.AddInt32( "nodeID", nID );
	}
	// Go through each selected item (i is < 0 if nothing selected).
	for (int i = m_pcListView->GetFirstSelected(); i >= 0 && i <= m_pcListView->GetLastSelected(); i++)
	{
		if( m_pcListView->IsSelected( i ) ) {
			os::ListViewStringRow *row = (os::ListViewStringRow *) m_pcListView->GetRow(i);
			os::String cStr = row->GetString(ID_INDEX);
			int nodeID = atoi(cStr.c_str());

			cMessage.AddInt32("nodeID", nodeID);

			DEBUG("ProgressWindow: sending Resume Message for node %d\n", nodeID);
		}
	}
	m_pcMainWindow->PostMessage(&cMessage, m_pcMainWindow); // Send the message to the main window
	return;
}


/** Send a cancel message for the current download to the main window.
 */
void ProgressWindow::SendCancelMessage( os::Message* pcMsg )
{
	os::Message cMessage( M_PW_CANCEL );
	int nID;
	if( pcMsg->FindInt32( "id", &nID ) == 0 ) {
		cMessage.AddInt32( "nodeID", nID );
	}
	// Go through each selected item (i is < 0 if nothing selected).
	for (int i = m_pcListView->GetFirstSelected(); i >= 0 && i <= m_pcListView->GetLastSelected(); i++)
	{
		if( m_pcListView->IsSelected( i ) ) {
			os::ListViewStringRow *row = (os::ListViewStringRow *) m_pcListView->GetRow(i);
			os::String cStr = row->GetString(ID_INDEX);
			int nodeID = atoi(cStr.c_str());

			cMessage.AddInt32("nodeID", nodeID);

			DEBUG("ProgressWindow: sending Cancel Message for node %d\n", nodeID);
		}
	}
	m_pcMainWindow->PostMessage(&cMessage, m_pcMainWindow); // Send the message to the main window
	return;
}


bool ProgressWindow::OkToQuit()
{
	os::Application::GetInstance()->PostMessage( os::M_QUIT );
	return( true );
}

/** \brief Creates a context menu for the given transfer and opens it at the given position.
 */
void ProgressWindow::OpenContextMenu( const os::Point& cPos, int nID )
{
	DEBUG( "OpenContextMenu( %.0f,%.0f,  %i )\n", cPos.x, cPos.y, nID );

	os::Message cMsg;
	cMsg.AddInt32( "id", nID );
	os::Menu* pcMenu = new os::Menu( os::Rect(), "context_menu", os::ITEMS_IN_COLUMN );
	cMsg.SetCode( M_PAUSE );
	pcMenu->AddItem( "Pause", new os::Message( cMsg ) );
	cMsg.SetCode( M_RESUME );
	pcMenu->AddItem( "Resume", new os::Message( cMsg ) );
	cMsg.SetCode( M_CANCEL );
	pcMenu->AddItem( "Cancel", new os::Message( cMsg ) );
	
	pcMenu->SetTargetForItems( this );
	pcMenu->Open( cPos );
	/* TODO: is the menu properly deleted? */
}


