#include <curl/curl.h>
#include <storage/file.h>
#include <util/messenger.h>
#include <util/datetime.h>

#include "queuenode.h"
#include "server.h"
#include "remotenode.h"
#include "messages.h"

/* Include: extern int ftpparse(struct ftpparse *,const char *,int); */
extern "C" {
#include "ftpparse.h"
};

#define DEBUG( arg... ) do { if( m_pcServer->GetSettings()->GetDebugMode() ) printf( arg ); } while(0)

/** \brief QueueNode Constructor.
 * Initializes the QueueNode storing all the data related to the node.
 */
QueueNode::QueueNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nType, int nID )
{
	m_pcServer = pcServer;
	
	
	m_zLocalPath = zLocalPath;
	m_zRemotePath = zRemotePath;
	m_nType = nType;
	m_nStatus = NODE_QUEUED;
	m_nID = nID;
	
	m_nTotalSize = -1;		/* -1 indicates 'not set yet' */
	m_nTotalDownloaded = -1;
	m_nTotalUploaded = -1;
}

/** \brief Set the necessary options in the curl handle.
 * The important parts are done in the subclasses.
 */
CURLcode QueueNode::AttachToHandle( CURL* pHandle )
{
	/* Call curl_easy_setopt() to set url, callbacks, etc. */
	/* We set some general options here; specifics like resume position etc will be set by subclasses. */

	/* Set callback functions */
	curl_easy_setopt( pHandle, CURLOPT_WRITEFUNCTION, QueueNode::WriteCallback );
	curl_easy_setopt( pHandle, CURLOPT_WRITEDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_READFUNCTION, QueueNode::ReadCallback );
	curl_easy_setopt( pHandle, CURLOPT_READDATA, this );
//	curl_easy_setopt( pHandle, CURLOPT_SEEKFUNCTION, QueueNode::SeekCallback );	/* Not available in older versions of libcurl */
//	curl_easy_setopt( pHandle, CURLOPT_SEEKDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_PROGRESSFUNCTION, QueueNode::ProgressBarCallback );
	curl_easy_setopt( pHandle, CURLOPT_PROGRESSDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_NOPROGRESS, 0 );
	
	return( CURLE_OK );
}


/** \brief Remove any settings from the curl handle set in AttachToHandle().
 */
CURLcode QueueNode::RemoveFromHandle( CURL* pHandle )
{
	curl_easy_setopt( pHandle, CURLOPT_URL, "" );
	curl_easy_setopt( pHandle, CURLOPT_WRITEDATA, NULL );
	curl_easy_setopt( pHandle, CURLOPT_READDATA, NULL );
//	curl_easy_setopt( pHandle, CURLOPT_SEEKDATA, NULL );
	curl_easy_setopt( pHandle, CURLOPT_PROGRESSDATA, NULL );
	
	return( CURLE_OK );
}

/** \brief QueueNode Default constructor.
 */
QueueNode::~QueueNode()
{
}


/** \brief Read Callback
 * The libcurl read callback.
 * Can't pass a virtual function to libcurl so we use these static wrappers.
 *
 * \param pBuf The location to read input data to.
 * \param nSize The amount of data to be read.
 * \param nMult The size in bytes of each data unit.
 * \param pCookie This contains the QueueNode* that we want to call the Read method on.
 *
 */
size_t QueueNode::ReadCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->Read( pBuf, nSize * nMult ) );
}


/** \brief Write Callback
 * The libcurl write callback.
 * Can't pass a virtual function to libcurl so we use these static wrappers.
 *
 * \todo Verify descriptions of callback parameters with libcurl documentation.
 *
 * \param pBuf The data that QueueNode will parse.
 * \param nSize The size of the data being passed to the QueueNode.
 * \param nMult The size of the data type being passed.
 * \param pCookie This contains the QueueNode* that we want to call the Write method on.
 *
 */
size_t QueueNode::WriteCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->Write( pBuf, nSize * nMult ) );
}

/** \brief Seek Callback
 * The libcurl seek callback.
 * Can't pass a virtual function to libcurl so we use these static wrappers.
 *
 * \param pCookie This contains the QueueNode* that we want to call the Seek method on.
 * \param nOffset How many bytes to offset into the file.
 * \param nOrigin The origin from which to start the offset.
 */
int QueueNode::SeekCallback( void* pCookie, curl_off_t nOffset, int nOrigin )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->Seek( nOffset, nOrigin ) );
}

/** \brief Progress Bar Callback
 * The libcurl progress bar. callback.
 * Can't pass a virtual function to libcurl so we use these static wrappers.
 *
 * \param pCookie The pointer to the QueueNode object to send the ProgreesBar update to.
 * \param fDownTotal Total data downloaded to be downloaded
 * \param fDownNow Total data downloaded so far.
 * \param fUpTotal Total data to be uploaded.
 * \param fUpNow Total data uploaded so far.
 */
int QueueNode::ProgressBarCallback( void* pCookie, double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->ProgressBar( fDownTotal, fDownNow, fUpTotal, fUpNow ) );
}

/***** DownloadNode implementation ******/

/** \brief DownloadNode Constructor. 
 * Sets the initial download file location to null.
 */
DownloadNode::DownloadNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID )
	: QueueNode( pcServer, zLocalPath, zRemotePath, NODE_DOWNLOAD, nID )
{
	m_pcFile = NULL;	/* This is only created when AttachToHandle() is first called. */
}

/** \brief DownloadNode Deconstructor.
 * This frees up the file pointer m_pcFile if necessary.
 */
DownloadNode::~DownloadNode()
{
	if( m_pcFile )
		delete( m_pcFile );
}

/** \brief Setup Function for Curl Transfer.
 * Configures the given CURL handle to perform this transfer, such as setting the url.
 */
CURLcode DownloadNode::AttachToHandle( CURL* pHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	if( m_pcFile == NULL ) {
		m_pcFile = new File( m_zLocalPath, O_WRONLY | O_CREAT | O_TRUNC );
	}
	
	QueueNode::AttachToHandle( pHandle );
	

	curl_off_t nOffset = m_pcFile->Seek( 0, SEEK_CUR );	/* Get current position in file, for if we are resuming */
	
	/* Set the url, set to download */
	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";
	zURL += m_zRemotePath;
	DEBUG( "DownloadNode: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_UPLOAD, 0 );
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, nOffset );
	
	return( CURLE_OK );
}

CURLcode DownloadNode::RemoveFromHandle( CURL* pHandle )
{
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
	return( CURLE_OK );
}

/** \brief Write to file.
 * This performs the actual writing of data to the file.
 *
 * \param pBuf The data to write to the file.
 * \param nSize The size of the data being written.
 *
 * \return The amount of data actually written.
 */
size_t DownloadNode::Write( void* pBuf, size_t nSize )
{
	size_t nResult = m_pcFile->Write( pBuf, nSize );
	m_pcFile->Flush();
	return( nResult );
}

/** \brief Performs a seek operation on the file.
 * Seek into a particular position in the file so curl can begin operating from there.
 *
 * \param nOffset The offset into the file to start operating at.
 * \param nOrigin The origin of the offset.
 *
 * \todo Make sure that File::Seek() returns the same as libcurl expects.
 */
int DownloadNode::Seek( curl_off_t nOffset, int nOrigin )
{
	return( m_pcFile->Seek( nOffset, nOrigin ) );

}

/** \brief Updates the progress bar.
 * This just outputs progress information to the terminal.
 *
 * \param fDownTotal Total bytes to be downloaded.
 * \param fDownNow Total bytes downloaded so far.
 * \param fUpTotal Total bytes to upload.
 * \param fUpNow Total bytes uploaded at the moment.
 */
int DownloadNode::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	SetDownloadedTotal( (size_t)fDownNow );
	SetTotalSize( (size_t)fDownTotal );
	m_pcServer->Updated( m_nID );
//	DEBUG( "Progress for download %s: %.1f%%\n", m_zRemotePath.c_str(), fDownNow/fDownTotal * 100 );
	return( 0 );
}

/****** UploadNode implementation ********/

/** \brief UploadNode Constructor. 
 * Sets the initial upload file location to null.
 */
UploadNode::UploadNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID )
	: QueueNode( pcServer, zLocalPath, zRemotePath, NODE_UPLOAD, nID )
{
	m_pcFile = NULL;	/* This is only created when AttachToHandle() is first called. */
}

/** \brief UploadNode Deconstructor.
 * Frees the pointer to the upload file.
 */
UploadNode::~UploadNode()
{
	if( m_pcFile )
		delete( m_pcFile );
}

/** \brief Setup Handle for Curl Transfer.
 * This does the setup for an upload curl handler.
 */
CURLcode UploadNode::AttachToHandle( CURL* pHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	if( m_pcFile == NULL ) {
		m_pcFile = new File( m_zLocalPath, O_RDONLY );
	}
	
	QueueNode::AttachToHandle( pHandle );
	

#if 1
	/* I don't think resuming uploads is this easy with ilbcurl :( */
	curl_off_t nOffset = m_pcFile->Seek( 0, SEEK_CUR );	/* Get current position in file, for if we are resuming */
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, nOffset );
#else
	m_pcFile->Seek( 0, SEEK_SET );
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
#endif

	curl_off_t nSize = m_pcFile->GetSize();
	curl_easy_setopt( pHandle, CURLOPT_INFILESIZE_LARGE, nSize );
	
	/* Set the url, set to upload */
	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";
	zURL += m_zRemotePath;
	DEBUG( "UploadNode: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_UPLOAD, 1 );
	
	return( CURLE_OK );
}

CURLcode UploadNode::RemoveFromHandle( CURL* pHandle )
{
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
	curl_easy_setopt( pHandle, CURLOPT_INFILESIZE_LARGE, 0 );
	curl_easy_setopt( pHandle, CURLOPT_UPLOAD, 0 );
	
	return( CURLE_OK );
}

/** \brief Read from file.
 * This performs the read from the file and returns the size read.
 *
 * \param pBuf A pointer to a location that the data can be read into.
 * \param nSize The amount of data to be read.
 */
size_t UploadNode::Read( void* pBuf, size_t nSize )
{
	return( m_pcFile->Read( pBuf, nSize ) );
}

/** \brief Seek into the upload file.
 * Seek into a particular position in the file so curl can begin operating from there.
 *
 * \param nOffset The offset into the file to start operating at.
 * \param nOrigin The origin of the offset.
 *
 * \todo Make sure that File::Seek() returns the same as libcurl expects.
 */
int UploadNode::Seek( curl_off_t nOffset, int nOrigin )
{
	return( m_pcFile->Seek( nOffset, nOrigin ) );
	
}

/** \brief Progress Bar Update.
 * This just outputs progress information to the terminal.
 *
 * \param fDownTotal Total bytes to be downloaded.
 * \param fDownNow Total bytes downloaded so far.
 * \param fUpTotal Total bytes to upload.
 * \param fUpNow Total bytes uploaded at the moment.
 */
int UploadNode::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	SetUploadedTotal( (size_t)fUpNow );
	SetTotalSize( (size_t)fUpTotal );
	m_pcServer->Updated( m_nID );
//	DEBUG( "Progress for upload %s: %.1f%%\n", m_zRemotePath.c_str(), (float)(100 * fUpNow/fUpTotal) );
	return( 0 );
}

/****** DirListNode implementation *******/

/** \brief Default constructor for DirListNode.
 * This initializes the target of the directory listing and makes note that
 * this is starting the listing.
 */
DirListNode::DirListNode( Server* pcServer, const String& zRemotePath, Handler* pcTarget, int nID )
	: QueueNode( pcServer, "", zRemotePath, NODE_DIRLIST, nID )
{
	m_pcTarget = pcTarget;
	m_bInitial = true;
}

/** \brief Default deconstructor for DirListNode.
 */
DirListNode::~DirListNode()
{
}

/** \brief Setup the Curl handle for the directory listing.
 */
CURLcode DirListNode::AttachToHandle( CURL* pHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	
	QueueNode::AttachToHandle( pHandle );
	
	String zURL = zServer;
	zURL += "/";
	zURL += m_zRemotePath;
	zURL += "/";			/* Make it clear to libcurl that we want a dir listing */
	

	/* Set the url, set to upload */
	DEBUG( "DirListNode: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_UPLOAD, 0 );
//	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
	
	return( CURLE_OK );
}

/** \brief Find the first occurrence of nChar in the first nLength bytes of pzStr. 
 * If not found, points to the first char *after* the end of the string.
 */
char* strnchr( char* pzStr, size_t nLength, int nChar )
{
	size_t nPos = 0;
	while( nPos < nLength && pzStr[nPos] != nChar && pzStr[nPos] != 0 ) ++nPos;
	return( pzStr + nPos );
}

/** \brief Write the data into the data buffer.
 * This "writes" the data coming from the remote server into the RemoteNode
 * objects. These are then passed to RemoteIconView to display to the user.
 *
 * \warning The RemoteNode vector that is sent is assumed to be deleted by the
 * RemoteIconView that calls for the update when it is done processing.
 *
 * \bug Generated nodes do not have proper permissions due to a deficiency in
 * how ftpparse operates.
 *
 * \todo The dirlist parsing might be better in Server, but we need to do it here because we need to save any fragments of the last line.
 *
 * \todo Fix up initial, final flags that are sent to the RemoteIconView. Maybe remove them. Currently the 'final' flag isn't being sent.
 */
size_t DirListNode::Write( void* pBuf, size_t nSize )
{
	char* pzBuf = (char*)pBuf;
	String zTmp( pzBuf, nSize );
//	DEBUG( "DirListNode::Write( %s, %i bytes, %sinitial ):\n%s\n------ end listing data ------\n", m_zRemotePath.c_str(), (int)nSize, m_bInitial?"":"not ", zTmp.c_str() );

	std::vector< RemoteNode >* pacList = new std::vector< RemoteNode >;	/* Will be deleted by the RemoteIconView when it's finished processing */

	struct ftpparse sData;
	memset( &sData, 0, sizeof( sData ) );
	
	size_t nTotalRead = 0; // Total bytes read.
	String zLine = m_zLastLineFragment; // Copy the last fragment read if we didn't read a complete line at the end last time.
	size_t nLineLength = m_zLastLineFragment.size(); // Initialize the current line length.
	char* pzLine = pzBuf;
	
	// Read in the next line up to and excluding the '\n'
	nTotalRead = strnchr( pzLine, nSize, '\n' ) - pzLine;
	nLineLength += nTotalRead;
	
	// Append the new text to the end of the line.
	zLine += String( pzLine, nTotalRead );
	
	/* Didn't find a \n; the line is incomplete so save it for later */
	if( nTotalRead == nSize )
	{
		m_zLastLineFragment = zLine;
		delete( pacList );
		return( nSize );
	}
	
	do
	{
//		DEBUG( "pzBuf: %x  pzLine: %x  length: %i     pzLine - pzBuf: %i\n", pzBuf, pzLine, nLineLength, pzLine - pzBuf );
//		DEBUG( "Line: '%s' length: %i\n", zLine.c_str(), nLineLength );

		// Attempt to parse the current line.
		int nResult;
		nResult = ftpparse( &sData, zLine.c_str(), nLineLength );
//		DEBUG( "Parse result: %i\n", nResult );

		if( nResult == 1 )	/* 1 indicates success */
		{
			// Create a new RemoteNode for the directory entry.
			RemoteNode cNode;
			cNode.m_zName = String( sData.name, sData.namelen );
			cNode.m_bIsDir = sData.flagtrycwd;
			cNode.m_nSize = sData.size;
			cNode.m_cTimestamp = DateTime( sData.mtime );
			cNode.m_zPath.Format( "%s/%s", m_zRemotePath.c_str(), cNode.m_zName.c_str() );
			/* ftpparse doesn't provide permissions! Guess we'll do without for now. */
			cNode.m_nPermissions = 0777;
//			DEBUG( "Parsed: name '%s', path '%s', size %i, %s, %s\n", cNode.m_zName.c_str(), cNode.m_zPath.c_str(), cNode.m_nSize, (cNode.m_bIsDir?"dir":"file"), cNode.m_cTimestamp.GetDate().c_str() );
			pacList->push_back( cNode );
		}
		else
		{
			String zTmp = String( pzLine, nLineLength );	/* Just for the next printf */
			printf( "ParseDirListing: ftpparse() failed on '%s' (%i bytes)\n", zTmp.c_str(), nLineLength );
		}

		nTotalRead++;	/* Skip the \n */
		pzLine = pzBuf + nTotalRead;
		nLineLength = strnchr( pzLine, nSize - nTotalRead, '\n' ) - pzLine;
		nTotalRead += nLineLength;
		zLine = String( pzLine, nLineLength );
		
		
		/* Save anything after the last \n (this will be empty if the pzBuf ends with \n) */
		if( nTotalRead >= nSize )
		{
			m_zLastLineFragment = zLine;
			break;
		}
	} while( 1 );
	
//	DEBUG( " nTotalRead: %i   nSize: %i   zLastLineFragment: '%s'\n", nTotalRead, nSize, m_zLastLineFragment.c_str() );
	
	if( pacList->size() == 0 )
	{
		/* We didn't sucessfully parse any entries! Clean up. */
		delete( pacList );
		return( nSize );
	}

	Message cMsg( M_REMOTE_DIRLISTING );
	cMsg.AddString( "path", m_zRemotePath );
	cMsg.AddPointer( "list", pacList );

	if( m_bInitial )
		cMsg.AddBool( "initial", true );	/* TODO: fix up initial, final flags. Maybe remove them. */
//	cMsg.AddBool( "final", true );
	
	Messenger cMessenger( m_pcTarget );
	cMessenger.SendMessage( &cMsg );
	
	m_bInitial = false;
	
	return( nSize );
}

/** \brief Update the Progress bar for the node.
 * This just outputs progress information to the terminal.
 *
 * \param fDownTotal Total bytes to be downloaded.
 * \param fDownNow Total bytes downloaded so far.
 * \param fUpTotal Total bytes to upload.
 * \param fUpNow Total bytes uploaded at the moment.
 */
int DirListNode::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
//	DEBUG( "Progress for dirlisting %s: %.1f out of %.1f\n", m_zRemotePath.c_str(), (float)fDownNow, (float)fDownTotal );
	return( 0 );
}

/**** CommandNode implementation *******/

CommandNode::CommandNode( Server* pcServer, const std::vector<String>& azCommands, int nID )
 : QueueNode( pcServer, "", "", NODE_COMMAND, nID )
{
	m_azCommands = azCommands;
	m_psCommands = NULL;
}

CommandNode::~CommandNode()
{
}

CURLcode CommandNode::AttachToHandle( CURL* pHandle )
{
	DEBUG( "CommandNode::AttachToHandle(): '%s'\n", m_azCommands.front().c_str() );
	QueueNode::AttachToHandle( pHandle );

	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";	/* Dummy path */
	
	/* Attach the commands to the handle */
	std::vector<String>::iterator i = m_azCommands.begin();
	for( ; i != m_azCommands.end(); i++ ) {
		m_psCommands = curl_slist_append( m_psCommands, i->c_str() );
	}

	curl_easy_setopt( pHandle, CURLOPT_QUOTE, m_psCommands );
	
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_NOBODY, 1 );

	curl_easy_setopt( pHandle, CURLOPT_NOPROGRESS, 1 );	/* Don't want progress notifications */

	return( CURLE_OK );
}

CURLcode CommandNode::RemoveFromHandle( CURL* pHandle )
{
	DEBUG( "CommandNode::RemoveFromHandle()\n" );
	QueueNode::RemoveFromHandle( pHandle );
	curl_easy_setopt( pHandle, CURLOPT_NOBODY, 0 );
	curl_easy_setopt( pHandle, CURLOPT_QUOTE, NULL );
	curl_easy_setopt( pHandle, CURLOPT_NOPROGRESS, 0 );
	
	curl_slist_free_all( m_psCommands );
	
	return( CURLE_OK );
}


/**** InitializationNode implementation ****/

InitializationNode::InitializationNode( Server* pcServer, int nID )
	: QueueNode( pcServer, "", "", NODE_COMMAND, nID )
{
}

InitializationNode::~InitializationNode()
{
}

CURLcode InitializationNode::AttachToHandle( CURL* pHandle )
{
	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";
	
	QueueNode::AttachToHandle( pHandle );
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_NOBODY, 1 );
	curl_easy_setopt( pHandle, CURLOPT_NOPROGRESS, 1 );
	
	/* This is a workaround for a libcurl bug - without using QUOTE, libcurl never considers the transfer finished. */
	struct curl_slist* pCommandList = curl_slist_append( NULL, "PWD" );
	curl_easy_setopt( pHandle, CURLOPT_QUOTE, pCommandList );
	
	return( CURLE_OK );
}

CURLcode InitializationNode::RemoveFromHandle( CURL* pHandle )
{
	/* Now the connection has been established, we can get the entry path */
	const char* pzEntryPath = NULL;
	curl_easy_getinfo( pHandle, CURLINFO_FTP_ENTRY_PATH, &pzEntryPath );
	if( pzEntryPath == NULL ) {
		printf( "InitializationNode: FTP_ENTRY_PATH returned NULL!\n" );
	} else {
		m_pcServer->m_pzEntryPath = pzEntryPath;
		DEBUG( "InitializationNode: Got entry path %s\n", pzEntryPath );
	}

//	curl_easy_setopt( pHandle, CURLOPT_URL, NULL );
	curl_easy_setopt( pHandle, CURLOPT_NOBODY, 0 );
	curl_easy_setopt( pHandle, CURLOPT_NOPROGRESS, 0 );
	curl_easy_setopt( pHandle, CURLOPT_QUOTE, NULL );

	QueueNode::RemoveFromHandle( pHandle );
	
	return( CURLE_OK );
}

