#ifndef __QUEUENODE_H__
#define __QUEUENODE_H__

#include <storage/file.h>
#include <util/string.h>
#include <util/handler.h>
#include <curl/curl.h>

#include <vector>

using namespace os;

// Prototype for the Server class.
class Server;

/* Node status enum */
enum {
	NODE_QUEUED = 1,
	NODE_RUNNING,
	NODE_PAUSED
};

enum {
	NODE_DOWNLOAD,
	NODE_UPLOAD,
	NODE_DIRLIST,
	NODE_COMMAND
};

/** \brief Class for storing nodes in a queue.
 * This class is used to store nodes in the queue of things to transfer.
 *
 */
class QueueNode
{
protected:
	QueueNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nType, int nID );

public:
	virtual ~QueueNode();	/* I want this to be private but can't be bothered debugging the compiler errors for now */
	
	/** Return the current status of the node.
	 * \todo Should this be here since it is a GUI access methods?
	 */
	int GetStatus()
	{
		return( m_nStatus );
	}
	
	/** \brief Return the type of the transaction.
	 * \todo Should this be here since it is a GUI access methods?
	 */
	int GetType()
	{
		return( m_nType );
	}
	
	/** \brief Return the current progress of the QueueNode's operation.
	 * \todo Currently this just returns 0.
	 * \todo Should this be here since it is a GUI access methods?
	 */
	float GetProgress()
	{
		return( 0 );
	}
	
	String GetRemotePath()
	{
		return( m_zRemotePath );
	}
	
	String GetLocalPath()
	{
		return( m_zLocalPath );
	}

	/** \brief Find how much was downloaded so far (in bytes). */
	ssize_t GetDownloadedTotal()
	{
		return m_nTotalDownloaded;
	}
	
	/** \brief Find how much was downloaded so far (in bytes). */
	ssize_t GetUploadedTotal()
	{
		return m_nTotalUploaded;
	}

	/** \brief Find the actual size of the file being transferred (in bytes). */
	ssize_t GetTotalSize()
	{
		return m_nTotalSize;
	}
	
private:

	/** \brief Read data into the provided buffer.
	 *
	 * \param pBuf Pointer to the buffer to read data into.
	 * \param nSize The amount of data to be read.
	 *
	 * \warning This is only a place holder and does nothing.
	 * It is meant to be subclassed by the different types of QueueNodes.
	 */
	virtual size_t Read( void* pBuf, size_t nSize )
	{
		printf( "Warning: QueueNode::Read() called!\n" ); 
		return( 0 );
	}
	
	/** \brief Write data from the provided buffer.
	 *
	 * \param pBuf The location of the data that should be read from to write information.
	 * \param nSize Size of the data that is in pBuf.
	 * \warning This is only a place holder and does nothing.
	 * It is meant to be subclassed by the different types of QueueNodes.
	 */
	virtual size_t Write( void* pBuf, size_t nSize )
	{
		printf( "Warning: QueueNode::Write() called!\n" );
		return( 0 );
	}

	/** \brief Seek into the data that is being manipulated.
	 * \param nOffset Amount to offset into the data being manipulated.
	 * \param nOrigin Where to start offsetting from.
	 * \warning This is only a place holder and does nothing.
	 * It is meant to be subclassed by the different types of QueueNodes.
	 */	
	virtual int Seek( curl_off_t nOffset, int nOrigin )
	{
		printf( "Warning: QueueNode::Seek() called!\n" );
		return( -1 );
	}
	
	/** \brief Update the progress bar for the QueueNode.
	 *
	 * \param fDownTotal Total data downloaded to be downloaded
	 * \param fDownNow Total data downloaded so far.
	 * \param fUpTotal Total data to be uploaded.
	 * \param fUpNow Total data uploaded so far.
	 *
	 * \warning This is only a place holder and does nothing.
	 * It is meant to be subclassed by the different types of QueueNodes.
	 */
	virtual int ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
	{
		printf( "Warning: QueueNode::ProgressBar() called!\n" );
		return( 0 );
	}

	static size_t ReadCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie );
	static size_t WriteCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie );
	static int SeekCallback( void* pCookie, curl_off_t nOffset, int nOrigin );
	static int ProgressBarCallback( void* pCookie, double fDownTotal, double fDownNow, double fUpTotal, double fUpNow );

protected:
	/** \brief Set how much was downloaded so far.
	 * \param uSize The amount of data downloaded so far (in bytes).
	 */
	void SetDownloadedTotal(ssize_t nSize)
	{
		m_nTotalDownloaded = nSize;
	}

	/** \brief Set how much was uploaded so far.
	 * \param uSize The amount of data uploaded so far (in bytes).
	 */
	void SetUploadedTotal(ssize_t nSize)
	{
		m_nTotalUploaded = nSize;
	}
	
	/** \brief Set the actual size of the file being downloaded.
	 * \param uSize The size of the file (in bytes).
	 */
	void SetTotalSize(ssize_t nSize)
	{
		m_nTotalSize = nSize;
	}

protected:
	virtual CURLcode AttachToHandle( CURL* pHandle );
	virtual CURLcode RemoveFromHandle( CURL* pHandle );

	/** \brief Server that this QueueNode belongs to. */
	Server* m_pcServer;

	/** \brief Local path of the file to be transfered. */
	String m_zLocalPath;
	
	/** \brief Remote path of the file to be transfered. */
	String m_zRemotePath;
	
	/** \brief Stores type of transfer this node represents.
	 * Could be: Upload, Download, Directory Listing, etc.
	 *
	 * See the enums in the header file.
	 *
	 * \todo Should this be an enum type?
	 */
	int m_nType;
	
	/** \brief Stores the status of the node.
	 * Should use one of the enums for Active or Queued.
	 *
	 * See the enums in the header file.
	 *
	 * \todo Should this be an enum type?
	 */
	int m_nStatus;
	
	/** \brief Our id for the Server's std::map. */
	int m_nID;
	
	/* Upload/download totals. */
	ssize_t m_nTotalSize;
	ssize_t m_nTotalDownloaded;
	ssize_t m_nTotalUploaded;
	
	friend class TransferThread;
};

/** \brief Download Queue Node.
 * This is a subclass of the QueueNode type that is used for
 * download requests in the queue.
 */
class DownloadNode : public QueueNode
{
public:
	DownloadNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID );
	~DownloadNode();

	
private:
	CURLcode AttachToHandle( CURL* pHandle );
	CURLcode RemoveFromHandle( CURL* pHandle );

	size_t Write( void* pBuf, size_t nSize );
	int Seek( curl_off_t nOffset, int nOrigin );
	int ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow );

	File* m_pcFile; /**< A pointer to the file that is being downloaded. */
	
	friend class TransferThread;
};

/** \brief Upload Queue Node.
 * This is a subclass of the QueueNode type that is used for
 * upload requests in the queue.
 */
class UploadNode : public QueueNode
{
public:

	UploadNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID );
	~UploadNode();

private:
	CURLcode AttachToHandle( CURL* pHandle );
	CURLcode RemoveFromHandle( CURL* pHandle );
	
	size_t Read( void* pBuf, size_t nSize );
	int Seek( curl_off_t nOffset, int nOrigin );
	int ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow );

	File* m_pcFile; /**< A pointer to the file that is being uploaded. */

	friend class TransferThread;
};

/** \brief Directory Listing Queue Node.
 * This is a subclass of the QueueNode type that is used for
 * directory listing requests in the queue.
 */
class DirListNode : public QueueNode
{
private:
	DirListNode( Server* pcServer, const String& zRemotePath, Handler* pcTarget, int nID );
	~DirListNode();
	CURLcode AttachToHandle( CURL* pHandle );
	
	size_t Write( void* pBuf, size_t nSize );
	int ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow );
	
	/** The target location to send the directory listing to. */
	Handler* m_pcTarget;
	
	/** Whether or not this is the initial phase of the directory listing download. */
	bool m_bInitial;
	
	/** \brief Stores anything after last line ending with '\n'.
	 * This is used for when the read operation did read up to the new line.
	 * It is used to continue reading from the output of Curl. */
	String m_zLastLineFragment;

	friend class TransferThread;
};

/** \brief A dummy node that contains commands we want to send to the server (eg 'delete').
 */
class CommandNode : public QueueNode
{
private:
	CommandNode( Server* pcServer, const std::vector<String>& azCommands, int nID );
	~CommandNode();
	
	CURLcode AttachToHandle( CURL* pHandle );
	CURLcode RemoveFromHandle( CURL* pHandle );
	
	std::vector<String> m_azCommands;
	
	struct curl_slist* m_psCommands;
	
	friend class TransferThread;
};

/** \brief A dummy node whose only purpose is to get the user's entry path.
 *  It calls curl_easy_getinfo() in RemoveFromHandle() to get the entry path.
 *  This is a bit of a hack but I can't think of a better way to do it.
 */
class InitializationNode : public QueueNode
{
private:
	InitializationNode( Server* pcServer, int nID );
	~InitializationNode();
	
	CURLcode AttachToHandle( CURL* pHandle );
	CURLcode RemoveFromHandle( CURL* pHandle );
	
	friend class TransferThread;
};

#endif	/* __QUEUENODE_H__ */

