#include "server.h"
#include "queuenode.h"
#include "remotenode.h"
#include "settings.h"
#include "messages.h"
#include "UpdateMsg.h"

#define DEBUG( arg... ) do { if( GetSettings()->GetDebugMode() ) printf( arg ); } while(0)


/** \brief Default constructor for the Server.
 * This initializes all the details about the server.
 *
 * \param zURL This is the string containing the parsed server URL to connect to.
 * \param pcTarget Pointer to the handler for server messages (ie. the window containing the server).
 * \param nParallelConnections Number of allowable parallel connections (simultaneous transfers).
 */
Server::Server( const String& zURL, os::Window* pcTarget, AppSettings* pcSettings )
{
	m_zURL = zURL;
	m_pcTarget = pcTarget;
	m_pcSettings = pcSettings;
	m_pzEntryPath = NULL;
	
	m_nNextID = 1;
	
	m_pcLock = new Locker( "server_lock" );
	
	m_pcThread = new TransferThread( this );
	m_pcThread->Start();
}

/** \brief Server Deconstructor.
 * This removes the server lock and quits. The m_pcThread is cleaned up
 * on its own when it receives the M_TT_CLOSE message.
 *
 * \todo Is it true that all other cleanup is done in TransferThread.
 * \todo I want this to be private but can't be bothered debugging the error messages yet.
 */
Server::~Server()
{
	delete( m_pcLock );
}

/** \brief Lock the Server.
 * This acts as a mutex lock for the Server so that the internal
 * data structures aren't corrupted due to concurrency issues.
 *
 * \warning Must call Lock() before accessing the node queue.
 */
int Server::Lock()
{
	return( m_pcLock->Lock() );
}

/** \brief Unlock the Server.
 * This unlocks the Server so that other items can operate on
 * the server.
 *
 * \warning This must be called by the function that called Lock() to
 * allow it to be used by other parts of the application.
 */
int Server::Unlock()
{
	return( m_pcLock->Unlock() );
}

/** \brief Returns the server address.
 * Returns the address part of the url, eg ftp://my.ftp.com
 */
String Server::GetServerAddress()
{
	return( m_zURL );
}

/** \brief Get the entry path of the server.
 *  Returns the entry path for the user's ftp session.
 *  This is the current directory immediately after logging on to the server (eg, user's home dir).
 *  The pointer is provided by libcurl.
 *  If libcurl hasn't provided the entry path yet, this returns NULL.
 */
const char* Server::GetEntryPath()
{
	return( m_pzEntryPath );
}

/** \brief Close all connections.
 * Stop all transfers, close the server connection, 
 * free curl handles, stop transfer thread, delete self.
 */
int Server::Close()
{
	m_pcThread->SendMessage( M_TT_CLOSE );
	return( 0 );
}

/** \brief Get Remote file.
 * Queue a download from the remote file to the local file.
 *
 * \param zRemoteSource Remote source relative to the ftp server, eg /remotedir/file
 * \param zLocalDest Local destination for the file.
 */
int Server::GetRemote( const String& zRemoteSource, const String& zLocalDest )
{
	DEBUG( "GetRemote( %s, %s )\n", zRemoteSource.c_str(), zLocalDest.c_str() );
	Message* pcMsg = new Message( M_TT_ADD );
	pcMsg->AddString( "localpath", zLocalDest );
	pcMsg->AddString( "remotepath", zRemoteSource );
	pcMsg->AddInt32( "type", NODE_DOWNLOAD );
	return( m_pcThread->SendMessage( pcMsg ) );
}

/** \brief Send a local file.
 * Queue an upload of the given local file to the given remote file. 
 *
 * \param zRemoteDest Remote destination relative to the ftp server, eg /remotedir/file
 * \param zLocalSource Local source where the local file is coming from.
 */
int Server::SendLocal( const String& zLocalSource, const String& zRemoteDest )
{
	DEBUG( "SendLocal( %s, %s )\n", zLocalSource.c_str(), zRemoteDest.c_str() );
	Message* pcMsg = new Message( M_TT_ADD );
	pcMsg->AddString( "localpath", zLocalSource );
	pcMsg->AddString( "remotepath", zRemoteDest );
	pcMsg->AddInt32( "type", NODE_UPLOAD );
	return( m_pcThread->SendMessage( pcMsg ) );
}

/** \brief Delete a remote file.
 * Tells the transfer thread to queue the given file for deletion.
 */
int Server::DeleteRemote( const String& zPath )
{
	Message* pcMsg = new Message( M_TT_DELETE );
	pcMsg->AddString( "remotepath", zPath );
	return( m_pcThread->SendMessage( pcMsg ) );
}

/** \brief Delete a remote directory.
 * Tells the transfer thread to queue the given dir for deletion.
 */
int Server::RemoveRemoteDir( const String& zPath )
{
	Message* pcMsg = new Message( M_TT_RMDIR );
	pcMsg->AddString( "remotepath", zPath );
	return( m_pcThread->SendMessage( pcMsg ) );
}

/** \brief Create a remote directory.
 * Tells the transfer thread to queue a request to create the given dir.
 */
int Server::CreateRemoteDir( const String& zPath )
{
	Message* pcMsg = new Message( M_TT_MKDIR );
	pcMsg->AddString( "remotepath", zPath );
	return( m_pcThread->SendMessage( pcMsg ) );
}

/** \brief Move/rename a remote file or dir.
 * Tells the transfer thread to queue a request to rename the given file or dir.
 */
int Server::MoveRemote( const String& zOldPath, const String& zNewPath )
{
	Message* pcMsg = new Message( M_TT_RENAME );
	pcMsg->AddString( "remotepath", zOldPath );
	pcMsg->AddString( "newpath", zNewPath );
	return( m_pcThread->SendMessage( pcMsg ) );
}


/** \brief Get Directory Listing.
 * Creates and sends a message to the m_pcThread that will cause it
 * to download a directory listing for the given path and generate the
 * corresponding RemoteNodes for each listing. The resulting nodes should
 * be sent to pcTarget.
 *
 * \param zPath The path to get a directory listing of.
 * \param pcTarget The handler for the resulting nodes.
 */
int Server::GetDirListing( const String& zPath, Handler* pcTarget )
{
	Message* pcMsg = new Message( M_TT_ADD );
	pcMsg->AddString( "remotepath", zPath );
	pcMsg->AddString( "localpath", "" );
	pcMsg->AddInt32( "type", NODE_DIRLIST );
	pcMsg->AddPointer( "target", pcTarget );
	return( m_pcThread->SendMessage( pcMsg ) );
}

/** \brief Pause Transfer of a given transfer thread.
 * This pauses the transfer with the given node id.
 *
 * \param nNodeID The node ID of the node to pause the transfer.
 */
int Server::PauseTransfer( int nNodeID )
{
	Message* pcMsg = new Message( M_TT_PAUSE );
	pcMsg->AddInt32( "id", nNodeID );

#ifdef DEBUG_PW
	DEBUG("Server: Received Pause for Node %d.\n", nNodeID);
#endif
	return( m_pcThread->SendMessage( pcMsg ) );
}

/** \brief Resume Transfer of a given transfer thread.
 * This resumes the transfer of the item with the given node id.
 *
 * \param nNodeID The node ID of the thread to resume transfering.
 */
int Server::ResumeTransfer( int nNodeID )
{
	Message* pcMsg = new Message( M_TT_RESUME );
	pcMsg->AddInt32( "id", nNodeID );
#ifdef DEBUG_PW
	DEBUG("Server: Received Resume for Node %d.\n", nNodeID);
#endif
	return( m_pcThread->SendMessage( pcMsg ) );
}

/** \brief Cancel Transfer of a given transfer thread.
 * This cancels the transfer of the item with the given node id.
 *
 * \param nNodeID The node ID of the thread to cancel transfering.
 */
int Server::CancelTransfer( int nNodeID )
{
	Message* pcMsg = new Message( M_TT_CANCEL );
	pcMsg->AddInt32( "id", nNodeID );
#ifdef DEBUG_PW
	DEBUG("Server: Received Cancel for Node %d.\n", nNodeID);
#endif
	return( m_pcThread->SendMessage( pcMsg ) );
}

/**** The following methods are only called from within the transfer thread ****/

/** \brief Setup Curl Handle.
 * This initializes the curl handler for the FTP connection.
 *
 * \param pHandle The pointer to the CURL handle.
 *
 * \todo This should be a virtual function that is called by the particular server
 * type. Currently it directly implements FTP style transfers which defeates the
 * purpose of having separate Server and FTPServer classes.
 */
CURLcode Server::_SetupCurlHandle( CURL* pHandle )
{
	curl_easy_setopt( pHandle, CURLOPT_VERBOSE, m_pcSettings->GetDebugMode() );	/* For testing */
	curl_easy_setopt( pHandle, CURLOPT_FTPPORT, m_pcSettings->GetPassive() ? NULL : "-" );
	curl_easy_setopt( pHandle, CURLOPT_FTP_RESPONSE_TIMEOUT, (long)20 );
	curl_easy_setopt( pHandle, CURLOPT_TIMEOUT, (long)100000 );
	/* That's it for now; other options will be set by subclasses or by QueueNode::AttachToHandle() */
	return( CURLE_OK );
}


/** \brief Updated server.
 * This is used to send a message to the main window signifying that some of the
 * downloads/uploads have been updated.
 * \param nID: The id of the node that has been updated.
 * Currently the nID parameter is ignored but a future version will do more fine-grained updating.
 */
void Server::Updated( int nID )
{
	m_pcTarget->PostMessage( new Message(M_QUEUE_NODES_UPDATED), m_pcTarget );
}


/** \brief Create update message to send to the progress window.
 */
Message* Server::CreateUpdateMessage()
{
	Message *update = NULL;
	std::map< int, QueueNode* >::iterator it;
	int length = 0;
	
	Lock();
	// First go through and only count the nodes that are a d/l or u/l 
	for (it = m_apcQueue.begin(); it != m_apcQueue.end(); it++)
	{
		// (*it).first - gets the key
		// (*it).second - gets the value
		switch(it->second->GetType())
		{
			case NODE_DOWNLOAD:
			case NODE_UPLOAD:
				length++;
				break;
			default:
				// ignore it
				break;
		}
	}
	
	if ( length > 0 )
	{
		// Fill in the update message.
		try
		{
			int nTotalDownloads = 0, nTotalUploads = 0, nTotalQueued = 0;
			update = new Message( M_PW_UPDATE );
			
			std::vector< QueueItem >* pacNodes = new std::vector< QueueItem >( length );
			/* This will be deleted by ProgressWindow when it receives the message */

			std::vector< QueueItem >::iterator j = pacNodes->begin();
			
			// Fill in the update message arrays.
			std::map< int, QueueNode* >::iterator it;
			for (it = m_apcQueue.begin(); it != m_apcQueue.end(); it++)
			{
				// (*it).first - gets the key
				// (*it).second - gets the value
				if( it->second->GetType() == NODE_DOWNLOAD || it->second->GetType() == NODE_UPLOAD )
				{
					// Copy over the file and server names
					j->zRemotePath = it->second->GetRemotePath();
					j->zLocalPath = it->second->GetLocalPath();

					// Fill in the data for the queue node.
					j->nID = it->first;
					j->eType = it->second->GetType();
					j->eStatus = it->second->GetStatus();

					j->nTotalDownload = it->second->GetDownloadedTotal();
					j->nTotalUpload = it->second->GetUploadedTotal();
					j->nTotalSize = it->second->GetTotalSize();
					
					if( it->second->GetStatus() == NODE_QUEUED ) {
						nTotalQueued++;
					} else if( it->second->GetType() == NODE_DOWNLOAD ) {
						nTotalDownloads++;
					} else {
						nTotalUploads++;
					}

					j++;
				}
			}
			
			update->AddInt32( "count", length );
			update->AddPointer( "data", pacNodes );
		}
		catch (...)
		{
			fprintf(stderr, "Internal Error: Server unable to send update to Progress Window!");
		}
	}
	else
	{
		try
		{
			update = new Message( M_PW_UPDATE );
			update->AddInt32( "count", 0 );
			update->AddPointer( "data", NULL );
		}
		catch(...)
		{
			// Just send back NULL;
			fprintf(stderr, "Internal Error: Server unable to send update to Progress Window!");
		}
	}
	Unlock();
	
	return update;
}


/**** FTPServer class ****
 * not implemented yet - all the functionality is in base Server class. 
 * Perhaps some functionality should be moved to FTPServer, or maybe we
 * can delete FTPServer.)
 *************************/


/** \brief FTPServer Constructor.
 *
 * \todo This should be filled in with things that are specific to FTP Server Connections.
 */
FTPServer::FTPServer( const String& zURL, Window* pcTarget, AppSettings* pcSettings )
	: Server( zURL, pcTarget, pcSettings )
{
}


/** \brief FTPServer Deconstructor.
 * Currently doesn't have to delete any allocated items.
 */
FTPServer::~FTPServer()
{
}

