#ifndef __TRANSFERTHREAD_H__
#define __TRANSFERTHREAD_H__

#include <util/thread.h>
#include <util/message.h>
#include <util/locker.h>
#include <list>
#include <queue>

#include <curl/curl.h>

using namespace os;

class Server;

class os::Handler;

/** CurlHandleStatus
 * Status numbers for CurlHandle_s */
enum CurlHandleStatus {
	HANDLE_IDLE = -1,
	HANDLE_INVALID = -2
};

/** \brief Structure for storing the pointer to the curl handle and its
 * current status.
 */
typedef struct {
	CURL*	m_pHandle; /**< Pointer to the curl handle. */
	int		m_nStatus;	/**< If positive, it is the index of a QueueNode in the Server's list; if negative it is one of the enum in CurlHandleStatus. */
} CurlHandle_s;

/** \brief Class for transfering items in a threaded manner.
 * This is the thread in which all the transfers are done via calls to 
 * curl_multi_perform().
 *
 * \todo The class needs to be cleaned up properly.
 */
class TransferThread : public Thread
{
public:
	TransferThread( Server* pcServer );
	~TransferThread();
	
	/* Main thread code is in Run() */
	int32 Run();
	
	/* Lock/unlock the message list so no-one will change it while we are using it */
	int Lock();
	int Unlock();
	
	/* Send a message to the transfer thread asking it to add, start, etc a given transfer */
	int SendMessage( int nCode );
	int SendMessage( Message* pcMsg );
	
private:
	void _Notify();		/* Send a signal to the thread to notify it that it has new messages */
	
	void _ProcessMessages();	/* Process the messages in the message list */
	
	void _Schedule();	/* Assign queued transfers from the queue to idle curl handles */
	
	void _Initialize();	/* Initialise the server connection - attach an InitializationNode to curl handles. */

	void _AddNode( int nType, const String& zLocalPath, const String& zRemotePath, Handler* pcTarget );
	void _PauseNode( int nID );
	void _ResumeNode( int nID );
	void _CancelNode( int nID );
	void _CleanupNode( int nNode );		/* Remove & delete a finished queue node */
	
	/* File management operations: delete, rename etc */
	void _DeleteFile( const String& zPath );
	void _RemoveDir( const String& zPath );
	void _MakeDir( const String& zPath );
	void _Rename( const String& zOldPath, const String& zNewPath );
	
	void _Close();	/* Clean up, delete objects and kill the thread */
	
	/** \brief Our server object */
	Server* m_pcServer;
	
	/* Inter-thread communication */
	int m_nReadPipe;  /**< Read pipe for communication. See pipe function in unistd.h. */
	int m_nWritePipe; /**< Write pipe for communication. See pipe function in unistd.h. */
	std::queue< Message* > m_apcMessages; /**< Stores the queue of messages for this thread.  \todo Could use std::pair< Transfer*, int > instead? */
	Locker* m_pcLock; /**< \brief A mutex lock for the thread. */
	
	/* libcurl stuff */
	CURLM* m_pCurlMultiHandle; /**< Our CURLM handle. Contains a pointer to our curl_multi_init() handle. \todo This needs better documentation. */
	std::vector< CurlHandle_s* > m_asHandles;	/**< List of all the curl handles we own, and the corresponding queue nodes. */
	
	friend class Server;
};



#endif	/* __TRANSFERTHREAD_H__ */






